from __future__ import annotations

from dataclasses import dataclass

from pytest_embrace.case import trickles
from pytest_embrace.codegen import CaseRender
from tests.conftest import ValidPythonAssertion
from tests.test_codegen.conftest import make_renderer_fixture


@dataclass
class TableCase:
    bottom: float = 0
    top: float = trickles()


CaseRenderUnderTest = CaseRender[TableCase]


renderer = make_renderer_fixture(TableCase, fixture_name="table")


def test_table_only(
    assert_valid_text_is: ValidPythonAssertion, renderer: CaseRenderUnderTest
) -> None:
    assert_valid_text_is(
        renderer.with_values(
            [
                TableCase(bottom=1.2, top=4.5),
                TableCase(bottom=4.0, top=99.9),
            ]
        ),
        """
        from pytest_embrace import CaseArtifact
        from tests.test_codegen.test_render_table_case import TableCase

        table = [
            TableCase(bottom=1.2, top=4.5),
            TableCase(bottom=4.0, top=99.9),
        ]


        def test(table: CaseArtifact[TableCase]) -> None:
            ...
        """,
    )


def test_table_and_module_level_attrs(
    assert_valid_text_is: ValidPythonAssertion, renderer: CaseRenderUnderTest
) -> None:
    """Handle the case where you want to render trickle-down attrs + `table`."""
    mixed_values = (
        TableCase(top=1),
        [
            TableCase(bottom=11.1),
            TableCase(bottom=4.0, top=99.9),
        ],
    )
    assert_valid_text_is(
        renderer.with_values(mixed_values),
        """
        from pytest_embrace import CaseArtifact
        from tests.test_codegen.test_render_table_case import TableCase

        top = 1

        table = [
            TableCase(bottom=11.1),
            TableCase(bottom=4.0, top=99.9),
        ]


        def test(table: CaseArtifact[TableCase]) -> None:
            ...
        """,
    )



from __future__ import annotations

import pytest

from .utils import make_autouse_conftest

_ = make_autouse_conftest(
    """
    from dataclasses import dataclass
    import pytest

    from pytest_embrace import  Embrace
    from pytest_embrace.case import CaseArtifact, trickles


    @dataclass
    class TrickleCase:
        snack: str
        beverage: str = trickles()
        ounces_of_beverage: int = trickles(no_override=True)


    embrace = Embrace(TrickleCase)


    @embrace.fixture
    def trickle_case(case: TrickleCase) -> None:
        pass"""
)


SIMPLE_TEST_MODULE = """
from conftest import TrickleCase

beverage = 'just water, thanks'
ounces_of_beverage = 16

table = [
    TrickleCase(snack='do you have any dates?'),
    TrickleCase(snack="I'm stuffed! No thanks.", beverage='espresso')
]

def test(trickle_case):
    bev = trickle_case.case.beverage
    snack = trickle_case.case.snack

    if bev == 'just water, thanks':
        assert snack == 'do you have any dates?'
    elif bev == 'espresso':
        assert snack == "I'm stuffed! No thanks."
    else:
        raise Exception("We should never get here.")
"""


def test_single(pytester: pytest.Pytester) -> None:
    pytester.makepyfile(SIMPLE_TEST_MODULE)
    outcome = pytester.runpytest()
    outcome.assert_outcomes(passed=2)


TEST_MODULE_ESCHEWS_TABLE = """
from conftest import TrickleCase

beverage = 'seriously just water, please.'
snack = '100,000 pistachios'
ounces_of_beverage = 1


def test(trickle_case):
    ...
"""


def test_no_table(pytester: pytest.Pytester) -> None:
    """You don't _have_ to use a table if you use trickles()."""
    pytester.makepyfile(TEST_MODULE_ESCHEWS_TABLE)
    outcome = pytester.runpytest()
    outcome.assert_outcomes(passed=1)


TEST_MODULE_ESCHEWS_TRICKLING = """
from conftest import TrickleCase


table = [
    TrickleCase(
        snack="do you have any dates?",
        beverage="just water, thanks",
        ounces_of_beverage=12,
    ),
    TrickleCase(
        snack="I'm stuffed! No thanks.",
        beverage="espresso",
        ounces_of_beverage=2,
    ),
]


def test(trickle_case):
    ...
"""


def test_no_trickle(pytester: pytest.Pytester) -> None:
    """You don't need to set the trickle attr if every test sets it."""
    pytester.makepyfile(TEST_MODULE_ESCHEWS_TRICKLING)
    outcome = pytester.runpytest()
    outcome.assert_outcomes(passed=2)


TEST_MODULE_MAKES_MISTAKE = """
from conftest import TrickleCase

ounces_of_beverage = 1000000

table = [
    TrickleCase(
        snack="do you have any dates?",
        beverage="just water, thanks",
    ),
    TrickleCase(
        snack="I'm stuffed! No thanks.",
        beverage="espresso",
        ounces_of_beverage=2,
    ),
]


def test(trickle_case):
    ...

"""


def test_cant_trickle_no_override(pytester: pytest.Pytester) -> None:
    """Ok... If you have a trickle that is `no_override`, you _can_
    set it on every test (like the previous test) but you _can't_
    set it at both levels."""
    pytester.makepyfile(TEST_MODULE_MAKES_MISTAKE)
    outcome = pytester.runpytest()
    outcome.assert_outcomes(errors=1)
    outcome.stdout.fnmatch_lines(
        # closing brackets are a problem for some reason
        "*Trickle-down attribute 'ounces_of_beverage cannot be overridden in table[1*"
    )


# beverage is not set anywhere
TEST_MODULE_UNSET_BOTH_PLACES = """
from conftest import TrickleCase


table = [
    TrickleCase(
        snack="do you have any dates?",
        ounces_of_beverage=1
    ),
]


def test(trickle_case):
    ...
"""


def test_cant_be_totally_missing(pytester: pytest.Pytester) -> None:
    """If a trickle attr isn't set at top level and also not in a case,
    that aint ok."""
    pytester.makepyfile(TEST_MODULE_UNSET_BOTH_PLACES)
    outcome = pytester.runpytest()
    outcome.assert_outcomes(errors=1)
    outcome.stdout.fnmatch_lines(
        [
            # closing brackets are a problem for some reason
            "*'beverage' is unset at the module level and in table[0*"
        ]
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import List

from pytest_embrace import trickles
from pytest_embrace.loader import load
from tests.conftest import ModuleInfoFactory


@dataclass
class AnnotatedWithTrickleCase:
    bar: int
    foo: List[str] = trickles()


def test_value_trickles_down(module_info_factory: ModuleInfoFactory) -> None:
    target = module_info_factory.build(
        AnnotatedWithTrickleCase,
        foo=["hey", "sup"],
        table=[
            AnnotatedWithTrickleCase(bar=55),
            AnnotatedWithTrickleCase(bar=66, foo=["woah there"]),
            AnnotatedWithTrickleCase(bar=66, foo=[]),
        ],
    )
    loaded = load(target)
    assert loaded[0].foo == ["hey", "sup"], (
        "When a table member isn't given an arg,"
        " it inherits it from the parent case."
    )
    assert loaded[1].foo == ["woah there"], "Table cases can override their parents"
    assert (
        loaded[2].foo == []
    ), "Table cases can override their parents with falsey values"



from __future__ import annotations

import pytest

from pytest_embrace.undot import undot_type_str


@pytest.mark.parametrize(
    "given, expected",
    [
        ("foo", "foo"),
        ("foo.bar", "bar"),
        # skipping this one since it's not really a valid type
        ("[foo.bar]", "[bar]"),
        ("foo.bar.baz.buzz.whizz", "whizz"),
        ("foo[baz.bar, bizz.bozz]", "foo[bar, bozz]"),
        ("typing.Callable[[str], typing.Any]", "Callable[[str], Any]"),
        (
            "typing.Callable[[fizz.buzz, fonz.bonz], buzz.bazz]",
            "Callable[[buzz, bonz], bazz]",
        ),
    ],
)
def test(given: str, expected: str) -> None:
    actual = undot_type_str(given)
    assert expected == actual



"""Test an implementation of a fixture that yields.
Ensure that the cleanup after the yield happens."""
from __future__ import annotations

import pytest

from .utils import make_autouse_conftest, make_test_run_outcome_fixture

_ = make_autouse_conftest(
    """
         from dataclasses import dataclass
         import pytest

         from pytest_embrace import Embrace
         from pytest_embrace.case import CaseArtifact


         @pytest.fixture
         def fix_yields() -> str:
             yield "Reginald"
             print("Cleaning up fixture")

         @dataclass
         class TheCase:
             name: str


         embrace = Embrace(TheCase)

         @embrace.fixture
         def cleanup_case(case: TheCase, fix_yields: str) -> None:
             assert case.name == fix_yields
             yield "yoooo"
             print("Cleaning up runner")
         """
)

single_test_outcome = make_test_run_outcome_fixture(
    """
    name = "Reginald"

    def test(cleanup_case):
        assert cleanup_case.actual_result == "yoooo"
        ...
    """,
    # '-s' is synonym for '--capture=no', causes print statements to render
    pytest_args=("-s",),
)


def test_cleanup_happens(single_test_outcome: pytest.RunResult) -> None:
    single_test_outcome.assert_outcomes(passed=1)
    output = single_test_outcome.stdout.str()
    assert "Cleaning up fixture" in output
    assert "Cleaning up runner" in output



# skip the future import because it breaks dataclass.fields()
# and turns each Field.type to a string!
# isort: dont-add-import: from __future__ import annotations


import sys
from collections import ChainMap
from dataclasses import dataclass, fields
from typing import Callable, Mapping, Union

from pytest_embrace.case import AttrInfo
from pytest_embrace.codegen import AttrRender


@dataclass
class XYZ:
    one: int
    two: Callable[[Union[ChainMap, int]], Mapping[str, int]]


_fields_lookup = {f.name: f for f in fields(XYZ)}


def test_render_import_statement_global() -> None:
    """Expect no import."""
    attr = _fields_lookup["one"]
    info = AttrInfo(attr)
    renderer = AttrRender(info)
    assert renderer.imports() == ""


def test_render_import_statement_builtins_and_third_party() -> None:
    """Expect every import.
    Note on Python < 3.9:
        3.9 and after alias typing.Callable, typing.Mapping, etc
        to collections.abc. So when solving imports the latter is 'correct'
        in newer Pythongs."""
    attr = _fields_lookup["two"]
    info = AttrInfo(attr)
    renderer = AttrRender(info)
    imports = renderer.imports()
    assert (
        imports
        == "\n".join(
            [
                "from collections import ChainMap",
                (
                    "from collections.abc import Callable, Mapping"
                    if sys.version_info >= (3, 9)
                    else "from typing import Callable, Mapping, Union"
                ),
                # Union is absent below 3.9 because it gets isort'd into the above
                "from typing import Union" if sys.version_info >= (3, 9) else "",
            ]
        ).strip()
    )



from __future__ import annotations

from textwrap import dedent
from types import ModuleType, SimpleNamespace
from typing import Any, Callable, Dict, Type, cast

import pytest

from pytest_embrace.case import CaseTypeInfo
from pytest_embrace.loader import ModuleInfo

pytest_plugins = "pytester"


class ModuleInfoFactory:
    @staticmethod
    def build(type: Type, *, fixture_name: str = "", **ns_kwargs: Any) -> ModuleInfo:
        default_kw: Dict[str, Any] = {}
        if "__name__" not in ns_kwargs:
            default_kw["__name__"] = "name_not_set"

        mod = cast(ModuleType, SimpleNamespace(**default_kw, **ns_kwargs))
        case_info_kwargs: Dict[str, Any] = (
            {} if fixture_name == "" else {"fixture_name": fixture_name}
        )
        return ModuleInfo(
            module=mod, case_type_info=CaseTypeInfo(type, **case_info_kwargs)
        )


@pytest.fixture
def module_info_factory() -> ModuleInfoFactory:
    return ModuleInfoFactory()


ValidPythonAssertion = Callable[[str, str], None]


@pytest.fixture
def assert_valid_text_is() -> ValidPythonAssertion:
    def do_assert(actual: str, expected: str) -> None:
        assert dedent(expected).lstrip("\n") == actual
        try:
            exec(actual, globals(), locals())  # check that it's valid
        except Exception:
            pytest.fail(f"Invalid python generated: {actual}")

    return do_assert



from __future__ import annotations

from dataclasses import dataclass

import pytest

from pytest_embrace.exc import CaseConfigurationError
from pytest_embrace.loader import load
from tests.conftest import ModuleInfoFactory


@dataclass
class ValidCase:
    attr1: str
    attr2: str
    attr3: int


def test_load_attributes(module_info_factory: ModuleInfoFactory) -> None:
    target = module_info_factory.build(ValidCase, attr1="hey", attr2="hello", attr3=444)
    (loaded,) = load(target)
    assert loaded == ValidCase("hey", "hello", 444)


def test_load_fail_on_missing(module_info_factory: ModuleInfoFactory) -> None:
    target = module_info_factory.build(ValidCase, attr1="yo")
    with pytest.raises(
        CaseConfigurationError, match="Only got attributes {'attr1': 'yo'}"
    ):
        load(target)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass

import pytest

from pytest_embrace.case import trickles
from pytest_embrace.codegen import (
    CaseRender,
    RenderImport,
    RenderModuleBody,
    RenderText,
)
from pytest_embrace.exc import CodeGenError
from tests.conftest import ValidPythonAssertion
from tests.test_codegen.conftest import make_renderer_fixture


@dataclass
class UnserializeableCase:
    collection: set[str] | None = None
    subcase: "UnserializeableCase" | None = None
    funcdef: Callable[..., object] | None = trickles(default=None)


CaseRenderUnderTest = CaseRender[UnserializeableCase]

renderer = make_renderer_fixture(UnserializeableCase, fixture_name="case")


def test_flat(
    assert_valid_text_is: ValidPythonAssertion, renderer: CaseRenderUnderTest
) -> None:
    values = UnserializeableCase(
        collection=RenderText('{"foo", "bar"}'),
        subcase=RenderText('UnserializeableCase({"haha"})'),
    )
    assert_valid_text_is(
        renderer.with_values(values, hinted=False),
        """
    from pytest_embrace import CaseArtifact
    from tests.test_codegen.test_render_render_custom_code import UnserializeableCase

    collection = {"foo", "bar"}
    subcase = UnserializeableCase({"haha"})


    def test(case: CaseArtifact[UnserializeableCase]) -> None:
        ...
        """,
    )


def test_render_function(
    assert_valid_text_is: ValidPythonAssertion, renderer: CaseRenderUnderTest
) -> None:
    """Cool one here. Demo special behavior for multi-line input & fuction defs."""
    values = UnserializeableCase(
        collection=RenderText("{}"),
        # this doesn't get rendered as an assignment
        funcdef=RenderText(
            """
            def funcdef(bar):
                return object()
            """
        ),
    )
    assert_valid_text_is(
        renderer.with_values(values, hinted=False),
        """
    from pytest_embrace import CaseArtifact
    from tests.test_codegen.test_render_render_custom_code import UnserializeableCase

    collection = {}


    def funcdef(bar):
        return object()


    def test(case: CaseArtifact[UnserializeableCase]) -> None:
        ...
        """,
    )


def test_render_function_name_mismatch(renderer: CaseRenderUnderTest) -> None:
    """Not allowed to define a function whose name doesn't match what's in the Case."""
    values = UnserializeableCase(
        collection=RenderText("{}"),
        funcdef=RenderText(
            """
            def uhoh(bar):
                return object()
            """
        ),
    )
    with pytest.raises(
        CodeGenError,
        match="Can't render function called 'uhoh' for arg 'funcdef'."
        " Rename the function to 'funcdef' and try again.",
    ):
        renderer.with_values(values, hinted=False),


def test_render_module_body(
    assert_valid_text_is: ValidPythonAssertion, renderer: CaseRenderUnderTest
) -> None:
    values = RenderModuleBody(
        RenderImport(from_="collections.abc", import_="Sequence"),
        RenderText("# color commentary!"),
        (
            UnserializeableCase(
                funcdef=RenderText(
                    """def funcdef(stuff: Sequence[int | float]) -> list[str]:
                           return [*map(str, stuff)]
                    """
                )
            ),
            [
                UnserializeableCase(collection=RenderText("{'foo', 'bar'}")),
                UnserializeableCase(collection=RenderText("{}")),
                UnserializeableCase(
                    subcase=RenderText('UnserializeableCase({"haha"})'),
                ),
            ],
        ),
    )
    assert_valid_text_is(
        renderer.with_values(values, hinted=False),
        """
    from collections.abc import Sequence

    from pytest_embrace import CaseArtifact
    from tests.test_codegen.test_render_render_custom_code import UnserializeableCase


    # color commentary!
    def funcdef(stuff: Sequence[int | float]) -> list[str]:
        return [*map(str, stuff)]


    table = [
        UnserializeableCase(collection={"foo", "bar"}),
        UnserializeableCase(collection={}),
        UnserializeableCase(subcase=UnserializeableCase({"haha"})),
    ]


    def test(case: CaseArtifact[UnserializeableCase]) -> None:
        ...
        """,
    )



from __future__ import annotations

from dataclasses import dataclass, field

from pytest_embrace.codegen import CaseRender
from tests.conftest import ValidPythonAssertion
from tests.test_codegen.conftest import make_renderer_fixture


@dataclass
class Case:
    number: int
    word: str
    items: list[str] = field(default_factory=list)


CaseRenderUnderTest = CaseRender[Case]

renderer = make_renderer_fixture(Case, fixture_name="case")


def test_skeleton(
    assert_valid_text_is: ValidPythonAssertion, renderer: CaseRenderUnderTest
) -> None:
    assert_valid_text_is(
        renderer.skeleton(),
        """
        from pytest_embrace import CaseArtifact
        from tests.test_codegen.test_render_flat_case import Case

        number: int
        word: str
        items: list[str]


        def test(case: CaseArtifact[Case]) -> None:
            ...
        """,
    )


def test_with_values(
    assert_valid_text_is: ValidPythonAssertion, renderer: CaseRenderUnderTest
) -> None:
    values = Case(number=5, word="hello", items=["yo"])
    assert_valid_text_is(
        renderer.with_values(values),
        """
        from pytest_embrace import CaseArtifact
        from tests.test_codegen.test_render_flat_case import Case

        number: int = 5
        word: str = "hello"
        items: list[str] = ["yo"]


        def test(case: CaseArtifact[Case]) -> None:
            ...
        """,
    )


def test_with_values_skips_defaults(
    assert_valid_text_is: ValidPythonAssertion, renderer: CaseRenderUnderTest
) -> None:
    values = Case(number=5, word="hello")  # when default is excluded, it never renders

    assert_valid_text_is(
        renderer.with_values(values),
        """
        from pytest_embrace import CaseArtifact
        from tests.test_codegen.test_render_flat_case import Case

        number: int = 5
        word: str = "hello"


        def test(case: CaseArtifact[Case]) -> None:
            ...
        """,
    )


def test_with_values_skips_defaults_even_when_passed(
    assert_valid_text_is: ValidPythonAssertion, renderer: CaseRenderUnderTest
) -> None:
    # empty items list is default
    # quirky, but it still won't render even though it was passed explicitly.
    values = Case(number=5, word="hello", items=[])

    assert_valid_text_is(
        renderer.with_values(values),
        """
        from pytest_embrace import CaseArtifact
        from tests.test_codegen.test_render_flat_case import Case

        number: int = 5
        word: str = "hello"


        def test(case: CaseArtifact[Case]) -> None:
            ...
        """,
    )



from __future__ import annotations

from textwrap import dedent


class EmbraceError(Exception):
    """Base for framework errors."""


class CaseConfigurationError(EmbraceError):
    """Raise at build/collection time."""


class TwoStepApiDeprecationError(EmbraceError):
    """Very specific error for methods deprecated after 3.0."""

    def __init__(self, *, deprecated_method: str) -> None:
        msg = f"""
        `{deprecated_method}` is deprecated in favor of `Embrace.fixture`.
        The new API unifies the old two-step process into one.
        and requires very minimal code changes.

        See migration guide in the docs:
            https://ainsleymcgrath.github.io/pytest-embrace/two-to-three/

        In future releases, this method will completely removed."""

        super().__init__(dedent(msg))


class CodeGenError(EmbraceError):
    """Raise when anything related to codegen goes awry."""



"""Assemble a test schema that includes mutable things.
Create 2 modules that implement it.
Module A imports `some_mutable_thing` from Module B and messes with it.
Verify that mutations during run of A do not spill over into B."""

from __future__ import annotations

import pytest

from tests.test_plugin.utils import make_autouse_conftest, make_test_run_outcome_fixture

_ = make_autouse_conftest(
    """
        from dataclasses import dataclass
        from typing import Dict, List
        from pytest_embrace import Embrace


        @dataclass
        class MutableCase:
            mut: Dict[str, int]


        mut_config = Embrace(MutableCase)


        @mut_config.fixture
        def mut_case(case: MutableCase):
            return None
        """
)


outcome = make_test_run_outcome_fixture(
    test_first="""
        mut = {"foo": 1}


        def test(mut_case):
            assert mut_case.case.mut == {"foo": 1}
        """,
    test_second="""
        from test_first import mut

        mut['bar'] = 6


        def test(mut_case):
            assert mut_case.case.mut == {"foo": 1, "bar": 6}
        """,
)


def test(outcome: pytest.RunResult) -> None:
    outcome.assert_outcomes(passed=2)



from __future__ import annotations

from dataclasses import dataclass

from pytest_embrace import RenderModuleBody, trickles
from pytest_embrace.codegen import CaseRender
from tests.conftest import ValidPythonAssertion
from tests.test_codegen.conftest import make_renderer_fixture


@dataclass
class ExpertCase:
    bottom: float
    top: float = trickles()


CaseRenderUnderTest = CaseRender[ExpertCase]


# @pytest.fixture
# def renderer(make_renderer: RendererMaker) -> CaseRender[ExpertCase]:
#     return make_renderer(ExpertCase, fixture_name="expert")


renderer = make_renderer_fixture(ExpertCase, fixture_name="expert")


def xtest_expert(
    assert_valid_text_is: ValidPythonAssertion, renderer: CaseRenderUnderTest
) -> None:
    assert_valid_text_is(
        renderer.with_values(
            RenderModuleBody(
                "# hehe begin w comment",
            )
        ),
        """
        from pytest_embrace import CaseArtifact
        from tests.test_codegen.test_render_expert import ExpertCase

        table = [
            TableCase(bottom=1.2, top=4.5),
            TableCase(bottom=4.0, top=99.9),
        ]


        def test(table: CaseArtifact[ExpertCase]) -> None:
            ...
        """,
    )



from __future__ import annotations

from typing import Any, Protocol, Type, TypeVar

import pytest

from pytest_embrace.case import CaseTypeInfo
from pytest_embrace.codegen import CaseRender

T = TypeVar("T")


class RendererMaker(Protocol):
    def __call__(self, case: Type[T], *, fixture_name: str) -> CaseRender[T]:
        ...


@pytest.fixture
def make_renderer() -> RendererMaker:
    def make(case: Type[T], *, fixture_name: str) -> CaseRender[T]:
        info: CaseTypeInfo[T] = CaseTypeInfo(case, fixture_name=fixture_name)
        return CaseRender(info)

    return make


def make_renderer_fixture(case_type: Type[T], *, fixture_name: str) -> Any:
    @pytest.fixture
    def fix(make_renderer: RendererMaker) -> Any:
        return make_renderer(case_type, fixture_name=fixture_name)

    return fix



from __future__ import annotations

import pytest

from tests.test_plugin.utils import make_autouse_conftest

_ = make_autouse_conftest(
    """
import sys

from dataclasses import dataclass

from pytest_embrace import Embrace, derive_from_filename


@dataclass
class AnnotatedCase:
    horse_type: str = derive_from_filename()


embrace_1 = Embrace(AnnotatedCase)


@embrace_1.fixture
def horse_case(case: AnnotatedCase) -> None:
    assert case.horse_type == 'palomino'
"""
)


TEST_FILE = """
def test(horse_case):
    ...
"""


def test_success(pytester: pytest.Pytester) -> None:
    dir = pytester.mkpydir("test_nested")
    file = dir / "test_palomino.py"
    file.touch()
    with file.open("w") as f:
        f.write(TEST_FILE)
    outcome = pytester.runpytest()
    outcome.assert_outcomes(passed=1)


def test_failure(pytester: pytest.Pytester) -> None:
    dir = pytester.mkpydir("test_nested")
    file = dir / "test_some_other_horse.py"
    file.touch()
    with file.open("w") as f:
        f.write(TEST_FILE)
    outcome = pytester.runpytest()
    outcome.assert_outcomes(errors=1)



# skip the future import because it breaks dataclass.fields()
# and turns each Field.type to a string!
# isort: dont-add-import: from __future__ import annotations

import sys
from collections import deque
from dataclasses import dataclass
from itertools import chain
from typing import Callable

from pytest_embrace.case import CaseTypeInfo
from pytest_embrace.codegen import CaseRender
from tests.conftest import ValidPythonAssertion


@dataclass
class GlobalsCase:
    x: float
    y: str


def test_to_text_globals(assert_valid_text_is: ValidPythonAssertion) -> None:

    target = CaseRender(CaseTypeInfo(GlobalsCase, fixture_name="the_case"))
    text = target.skeleton()
    assert_valid_text_is(
        text,
        """
        from pytest_embrace import CaseArtifact
        from tests.test_case.test_to_text import GlobalsCase

        x: float
        y: str


        def test(the_case: CaseArtifact[GlobalsCase]) -> None:
            ...
        """,
    )


@dataclass
class SomeCaseWithImports:
    foo: Callable[[deque], chain]


def test_to_text_builtins(assert_valid_text_is: ValidPythonAssertion) -> None:

    target = CaseRender(CaseTypeInfo(SomeCaseWithImports, fixture_name="the_case"))
    assert_valid_text_is(
        target.skeleton(),
        """
        from collections import deque
        from collections.abc import Callable
        from itertools import chain

        from pytest_embrace import CaseArtifact
        from tests.test_case.test_to_text import SomeCaseWithImports

        foo: Callable[[deque], chain]


        def test(the_case: CaseArtifact[SomeCaseWithImports]) -> None:
            ...
        """
        if sys.version_info >= (3, 9)
        else (
            """
            from collections import deque
            from itertools import chain
            from typing import Callable

            from pytest_embrace import CaseArtifact
            from tests.test_case.test_to_text import SomeCaseWithImports

            foo: Callable[[deque], chain]


            def test(the_case: CaseArtifact[SomeCaseWithImports]) -> None:
                ...
            """
        ),
    )



from __future__ import annotations

import pytest

from .utils import generated_module_stdout_factory, make_autouse_conftest

_ = make_autouse_conftest(
    """
from dataclasses import dataclass
import pytest

from pytest_embrace import Embrace
from pytest_embrace.case import CaseArtifact


@pytest.fixture
def fix() -> str:
    return "hey"


@dataclass
class MyCase1:
    name: str


embrace_1 = Embrace(MyCase1)


@embrace_1.fixture
def fix_1(case: MyCase1, fix: str) -> None:
    assert case.name == fix


@dataclass
class MyCase2:
    num: int
    foo: str


embrace_2 = Embrace(MyCase2)


@embrace_2.fixture
def fix_2(case: MyCase2, fix: str) -> None:
    assert len(case.num) >= len(fix)
    assert case.foo != 'papaya'"""
)


@pytest.mark.parametrize(
    "fixture_name, matchlines",
    [
        (
            "buzzz",
            # XXX maybe some leaky state? other fixtures are called out in this output
            # (thus the glob before fix_1)
            ["*No such fixture 'buzzz'. Your options are *'fix_1', 'fix_2'*"],
        ),
        (
            "fix_1",
            generated_module_stdout_factory(
                "name: str", fixture="fix_1", case_type="MyCase1"
            ),
        ),
        (
            "fix_2",
            generated_module_stdout_factory(
                "num: int", "foo: str", fixture="fix_2", case_type="MyCase2"
            ),
        ),
    ],
)
def test_gen_opt_can_find(
    fixture_name: str, matchlines: str, pytester: pytest.Pytester
) -> None:
    """Looking for the type hints that show up in generated modules."""
    outcome = pytester.runpytest(f"--embrace={fixture_name}")
    outcome.stdout.fnmatch_lines(matchlines, consecutive=True)



from __future__ import annotations

import sys

import pytest

from .utils import generated_module_stdout_factory, make_autouse_conftest

_ = make_autouse_conftest(
    """
import sys

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

from dataclasses import dataclass

from pytest_embrace import Embrace


@dataclass
class AnnotatedCase:
    name: Annotated[str, 'This is ... the name.']


embrace_1 = Embrace(AnnotatedCase)


@embrace_1.fixture
def anno_case(case: AnnotatedCase, fix: str) -> None:
    assert case.name == fix
"""
)


@pytest.mark.skipif(
    sys.version_info < (3, 9),
    reason="get_type(Annotated[...]) returns nothing in 3.8"
    " so this feature can't work at all",
)
def test_comment_annotated_attr(pytester: pytest.Pytester) -> None:
    outcome = pytester.runpytest("--embrace=anno_case")
    matchlines = generated_module_stdout_factory(
        "name: str  # This is ... the name.",
        fixture="anno_case",
        case_type="AnnotatedCase",
    )
    outcome.stdout.fnmatch_lines(matchlines, consecutive=True)



from __future__ import annotations

import json
import re
from collections import defaultdict
from collections.abc import Iterator
from dataclasses import asdict
from inspect import getmodule
from operator import itemgetter
from textwrap import dedent
from typing import Any, Generic, List, Tuple, Type, Union, cast, get_args, get_origin

import isort
from black import format_str
from black.mode import Mode
from pyperclip import sys

from pytest_embrace.case import AttrInfo, CaseType, CaseTypeInfo, Trickle
from pytest_embrace.embrace import CaseTypeRegistry
from pytest_embrace.exc import CodeGenError
from pytest_embrace.undot import undot_type_str


class CodeGenManager:
    fixture_name: str
    generator_name: str | None
    generator_kwargs: dict[str, Any]
    case_type: CaseTypeInfo | None

    def __init__(self, directive: str, *, registry: CaseTypeRegistry[CaseTypeInfo]):
        # identifier goes <fixture_name>[:<generator_name>]
        identifier, *key_value_pairs = directive.split(" ")
        try:
            self.fixture_name, self.generator_name = identifier.split(":", maxsplit=1)
        except ValueError:
            self.fixture_name = identifier
            self.generator_name = None

        self.generator_kwargs: Any = {}  # Any to quell concerns about **kwargs
        for pair in key_value_pairs:
            key, value = pair.split("=", maxsplit=1)
            try:
                self.generator_kwargs[key] = json.loads(value)
            except json.JSONDecodeError as e:
                # a string is expected to fail when it doesn't come in quoted.
                # only fail on other invalid json
                if not isinstance(value, str):
                    raise CodeGenError(
                        f"Generator '{self.generator_name}' supplied with"
                        f"non-serializeable value '{value}' for argument {key}"
                    ) from e

                # strings don't load
                self.generator_kwargs[key] = value

        self.case_type = registry.get(self.fixture_name)

    def render(self) -> str:
        if self.case_type is None:
            raise TypeError("No case found, can't render.")

        renderer = CaseRender(self.case_type)

        if self.generator_name is None:
            return renderer.skeleton()

        try:
            generator = self.case_type.generators[self.generator_name]
        except KeyError as e:
            raise TypeError(f"No custom generator named {self.generator_name}") from e

        src = generator(**self.generator_kwargs)

        if (
            isinstance(src, self.case_type.type)
            or isinstance(src, list)
            or (
                # two-item iterable where [0] is Case and rest is list[Case]
                len(src) == 2
                and isinstance(src[0], self.case_type.type)
                and isinstance(src[1], list)
            )
        ):
            return renderer.with_values(src, hinted=False)
        else:
            raise CodeGenError(
                f"Invalid return type from generator {self.generator_name}: {type(src)}"
            )


class CaseRender(Generic[CaseType]):
    funcdef_pattern = re.compile(r"\n{0,}def\s+(?P<funcname>\w+)\(.+\).*:")

    def __init__(self, src: CaseTypeInfo[CaseType]):
        self.src = src
        self.attr_render = {
            name: AttrRender(info) for name, info in self.src.type_attrs.items()
        }

    def with_values(
        self,
        values: CaseType
        | tuple[CaseType, list[CaseType]]
        | list[CaseType]
        | RenderModuleBodyValue,
        hinted: bool = True,
        no_body: bool = False,
    ) -> str:
        if isinstance(values, list):
            out = self._with_values_from_list(values)
        elif isinstance(values, RenderModuleBodyValue):
            out = self._expert(values)
        elif isinstance(values, tuple):
            # python 3.8 can't deal with list[] or | union in cast() calls
            # (even with future import)
            out = self._with_mixed_values(cast(Tuple[CaseType, List[CaseType]], values))
        else:
            out = self._with_values_from_case(values, hinted=hinted)
        if no_body:
            return out
        return self.module_text(body=out)

    def module_text(self, *, body: str = "") -> str:
        text = (
            f"{self.imports()}\n{body}\n"
            f"def test({self.src.fixture_name}: CaseArtifact[{self.src.type_name}])"
            " -> None: ..."
        )
        return format_str(isort.code(text, profile="black"), mode=Mode())

    def imports(self) -> str:
        source = getmodule(self.src.type)
        case_import = (
            RenderImport(mod_name, self.src.type_name)
            if (mod_name := getattr(source, "__name__", None)) is not None
            else ""
        )
        dependency_imports = "\n".join(
            render.imports() for render in self.attr_render.values()
        )
        always_import = "from pytest_embrace import CaseArtifact\n"
        return f"{always_import}\n{case_import}\n{dependency_imports}"

    def skeleton(self) -> str:
        module_attr_hints = "\n".join(attr.hint() for attr in self.attr_render.values())
        return self.module_text(body=module_attr_hints)

    def _with_values_from_case(self, case: CaseType, hinted: bool = True) -> str:
        assignments: list[str] = []
        values = asdict(case)
        for name, attr in self.src.type_attrs.items():
            value = values[name]
            if callable(attr.dc_field.default_factory):
                default = attr.dc_field.default_factory()
            else:
                default = attr.dc_field.default

            # would've expected value == default to be true w/ trickles() but appaz not!
            if value == default or isinstance(value, Trickle):
                continue

            if isinstance(value, RenderValue):
                if match := self.funcdef_pattern.match(str(value)):
                    funcname = match["funcname"]
                    if funcname != name:
                        raise CodeGenError(
                            f"Can't render function called '{funcname}'"
                            f" for arg '{name}'. Rename the function to '{name}'"
                            " and try again."
                        )
                    assignments.append(str(value))
                    continue

                assignments.append(
                    self.attr_render[name].assignment(value, hinted=False)
                )
                continue

            assignments.append(self.attr_render[name].assignment(value, hinted=hinted))

        return "\n".join(assignments)

    def _with_values_from_list(self, values: list[CaseType]) -> str:
        typical = values[0]
        text = ", ".join(
            typical.__class__.__name__
            + "("
            + self._with_values_from_case(v, hinted=False).replace("\n", ",")
            + ")"
            for v in values
        )
        # trailing comma always
        return f"table = [{text},]"

    def _with_mixed_values(self, values: tuple[CaseType, list[CaseType]]) -> str:
        module_attrs, table = values
        header = self._with_values_from_case(module_attrs, hinted=False)
        table_content = self._with_values_from_list(table)
        return f"{header}\n\n{table_content}"

    def _expert(self, body: RenderModuleBodyValue) -> str:
        out: list[str] = []
        for item in body.contents:
            if isinstance(item, RenderValue):
                out.append(str(item))
                continue
            out.append(self.with_values(item, no_body=True))

        return "\n".join(out)


class AttrRender:
    def __init__(self, src: AttrInfo) -> None:
        self.src = src

    def imports(self) -> str:
        lookup: dict[str, list[str]] = defaultdict(list)
        for typ in _unnest_generics(self.src.dc_field.type):
            modname = getattr(typ, "__module__", "")
            if modname == "builtins" or modname == "":
                continue
            if sys.version_info < (3, 9) and "collections.abc" in modname:
                # from 3.9 and onwards, things like Mapping and Callable
                # are aliases of collections.abc types.
                # before then, the stuff from collections.abc was not subscriptable,
                # so here we are.
                modname = "typing"
            own_name: str = getattr(
                typ,
                "__name__",
                getattr(
                    typ,
                    # as in the case of typing.Union in <=3.9,
                    # whose __class__ is a _SpecialForm
                    "_name",
                    getattr(typ.__class__, "__name__", ""),
                ),
            )
            if own_name == "Annotated":
                continue
            lookup[modname].append(own_name)

        return "\n".join(
            f"from {modname} import {', '.join(sorted(target))}"
            for modname, target in sorted(lookup.items(), key=itemgetter(0))
        )

    def hint(self) -> str:
        typ = (
            self.src.dc_field.type
            if self.src.annotations is None
            else self.src.annotations.type
        )
        type_str = undot_type_str(_stringify_type(typ))
        # get rid of dots and assume imports are solved
        type_str_parts = [t.lstrip("[").rstrip("]") for t in type_str.split(".")]
        if len(type_str_parts) > 1:
            type_str = type_str_parts[-1]
        comment = next(
            (
                v
                for v in getattr(self.src.annotations, "annotations", [])
                if isinstance(v, str)
            ),
            None,
        )
        text = f"{self.src.name}: {type_str}"
        if comment is not None:
            text += f"  # {comment}"
        return text

    def assignment(self, value: str | RenderValue, hinted: bool = True) -> str:
        return (
            f"{self.hint() if hinted else self.src.name}"
            f" = {value if isinstance(value, RenderValue) else repr(value)}"
        )


def _stringify_type(type: Type) -> str:
    return str(type) if not repr(type).startswith("<class") else type.__name__


def _unnest_generics(type: Type | list[Type]) -> Iterator[Type]:
    if isinstance(type, list):  # such as the first arg to Callable[]
        for member in type:
            yield from _unnest_generics(member)
        return

    generic_args = get_args(type)
    if generic_args == tuple():
        yield type
    elif (origin := get_origin(type)) is not None:
        yield origin

    for arg in generic_args:
        yield from _unnest_generics(arg)


def RenderText(text: str) -> Any:
    """Interpolate any text as Python into a @generator return."""
    return RenderValue(text)


def RenderImport(from_: str, import_: str, as_: str = "") -> Any:
    alias = as_ if as_ == "" else f" as {as_}"
    return RenderValue(f"from {from_} import {import_}{alias}")


class RenderValue:
    """Stand-in as something distinctly identifiable when using Render()."""

    def __init__(self, content: str) -> None:
        self.content = content

    def __str__(self) -> str:
        return dedent(self.content)

    def __repr__(self) -> str:
        return str(self)


# python 3.8 can't deal with list[] or | union in aliases
# (even with future import)
ModuleBodyContents = Union[List[CaseType], CaseType, RenderValue]


def RenderModuleBody(*contents: ModuleBodyContents) -> Any:
    return RenderModuleBodyValue(*contents)


class RenderModuleBodyValue:
    """Everything after the solved imports and before the `test()` function def.
    Given contents are rendered all together, separated by newlines."""

    def __init__(self, *contents: ModuleBodyContents) -> None:
        self.contents: list[ModuleBodyContents] = [*contents]



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict

import pytest

from pytest_embrace.exc import CaseConfigurationError
from pytest_embrace.loader import load
from tests.conftest import ModuleInfoFactory


@dataclass
class BuiltinsAttrsCase:
    string: str
    integer: int
    dictionary: Dict[str, str]


def test_happy(module_info_factory: ModuleInfoFactory) -> None:
    target = module_info_factory.build(
        BuiltinsAttrsCase, string="string", integer=5, dictionary={}
    )
    (loaded,) = load(target)
    expected = BuiltinsAttrsCase(string="string", integer=5, dictionary={})
    assert expected == loaded


def test_multiple_wrong_types(module_info_factory: ModuleInfoFactory) -> None:
    target = module_info_factory.build(
        BuiltinsAttrsCase, string="string", integer={}, dictionary=["nope."]
    )
    with pytest.raises(CaseConfigurationError, match="2 invalid attr values"):
        load(target)


def test_table_wrong_types(module_info_factory: ModuleInfoFactory) -> None:
    target = module_info_factory.build(
        BuiltinsAttrsCase,
        foo=["hey", "sup"],
        table=[
            BuiltinsAttrsCase(string="string", integer=5, dictionary={}),
            BuiltinsAttrsCase(string=555.555, integer=1, dictionary={}),  # type: ignore
        ],
    )
    with pytest.raises(CaseConfigurationError):
        load(target)



from __future__ import annotations

from textwrap import dedent
from typing import Any, Callable, List, Tuple

import pytest

TestFuncWithPytester = Callable[[pytest.Pytester], None]


def make_conftest(text: str, *, autose: bool = False) -> TestFuncWithPytester:
    @pytest.fixture(autouse=autose)
    def autouse_conftest(pytester: pytest.Pytester) -> None:
        pytester.makeconftest(dedent(text))

    return autouse_conftest


def make_autouse_conftest(text: str) -> Callable[[pytest.Pytester], None]:
    return make_conftest(text, autose=True)


def make_test_run_outcome_fixture(
    *args: Any, pytest_args: Tuple[Any, ...] = (), **kwargs: Any
) -> Callable[[pytest.Pytester], pytest.RunResult]:
    """For the common case of:
        creating a python file (or files)
        immediately running pytest,
        directly using the Outcome.
    If other fixtures or complex setup are needed, this is not for you.
    Kwargs become filenames with a .py added."""

    @pytest.fixture
    def fixture(pytester: pytest.Pytester) -> pytest.RunResult:
        pytester.makepyfile(*args, **kwargs)

        outcome = pytester.runpytest(*pytest_args)
        return outcome

    return fixture


def generated_module_stdout_factory(
    *type_hints: str,
    fixture: str,
    case_type: str,
    imports: str = "from pytest_embrace import CaseArtifact",
    typical_table_kwargs: List[str] | None = None,
) -> List[str]:

    typical_table_kwargs = (
        []
        if typical_table_kwargs is None
        else [
            "",
            "",
            "# table = [",
            f"#    {case_type}(",
            *[f"#        {kwarg}=..." for kwarg in typical_table_kwargs],
            "#     )",
            "# ]",
        ]
    )

    return [
        "Copying the following output to your clipboard:",
        "",
        *imports.split("\n"),
        "",
        f"from conftest import {case_type}",
        "",
        *type_hints,
        *typical_table_kwargs,
        "",
        "",
        f"def test({fixture}: CaseArtifact[{case_type}]) -> None:",
        "    ...",
    ]



from __future__ import annotations

from itertools import takewhile

DOT = "."
OPEN_GENERIC = "["
CLOSE_GENERIC = "]"
COMMA_SPACE = ", "
DELIMETERS = {OPEN_GENERIC, CLOSE_GENERIC, COMMA_SPACE}
SENTINEL = "~~~TERMINATE~~~"


def _dedot(string: str) -> str:
    return COMMA_SPACE.join(part.split(DOT)[-1] for part in string.split(COMMA_SPACE))


def undot_type_str(type_str: str) -> str:
    it = iter(type_str)

    while True:
        char = next(it, SENTINEL)
        if char == SENTINEL:
            return ""
        elif char not in DELIMETERS:
            word = _dedot(char + "".join(takewhile(lambda x: x not in DELIMETERS, it)))
            rest = type_str.split(word, maxsplit=1)[-1]
            return f"{word}{undot_type_str(rest)}"
        elif char == OPEN_GENERIC:
            inner_word = "".join(takewhile(lambda x: x != CLOSE_GENERIC, it))
            rest = type_str.split(inner_word, maxsplit=1)[-1]
            # for Callable[] the first generic arg is a [], so we need to check for that
            special_case = OPEN_GENERIC if inner_word.startswith(OPEN_GENERIC) else ""

            return (
                f"{char}{special_case}{_dedot(inner_word)}{undot_type_str(rest)}"
            ).replace(
                "[[[", "[["  # eye-bleedingly ugly solution to bug with unary
            )  # callables somehow gaining an extra bracket :(
        elif char == CLOSE_GENERIC:
            rest = type_str.split(char, maxsplit=1)[-1]
            word = _dedot(char + "".join(takewhile(lambda x: x not in DELIMETERS, it)))
            return f"{char}{_dedot(rest)}"



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Type, get_args

# class Prompt:
#     """When tests are generated via `--embrace`, prompt the user for this value."""


@dataclass
class AnnotationInfo:
    type: Any
    annotations: List[Any]


AnnotationMap = Dict[str, AnnotationInfo]


def get_pep_593_values(cls: Type) -> AnnotationMap:
    """Map string names of attributes to their annotation metadata."""
    out: AnnotationMap = {}
    for k, v in cls.__annotations__.items():
        attr_name = getattr(v, "__name__", getattr(v.__class__, "__name__", ""))
        if attr_name not in {"Annotated", "_AnnotatedAlias"}:
            continue
        typ, *annotations = get_args(v)
        out[k] = AnnotationInfo(typ, annotations)
    return out



from __future__ import annotations

import pytest

from ..utils import generated_module_stdout_factory, make_autouse_conftest

_ = make_autouse_conftest(
    """
from dataclasses import dataclass
import pytest

from pytest_embrace import Embrace
from pytest_embrace.case import CaseArtifact


@dataclass
class MyCase:
    name: str


embrace = Embrace(MyCase)


@embrace.fixture
def my_case(case: MyCase) -> None:
    pass


@embrace.generator
def dynamic_name(name: str):
    return MyCase(name=f"Dynamic {name.title()}!")
"""
)


def test_dynamic_name_generator(pytester: pytest.Pytester) -> None:
    outcome = pytester.runpytest("--embrace=my_case:dynamic_name name=pieRre")
    outcome.stdout.fnmatch_lines(
        generated_module_stdout_factory(
            'name = "Dynamic Pierre!"',
            case_type="MyCase",
            fixture="my_case",
        ),
    )



# skip the future import because it breaks dataclass.fields()
# and turns each Field.type to a string!
# isort: dont-add-import: from __future__ import annotations

import re
import sys
from dataclasses import _MISSING_TYPE, MISSING, Field, dataclass, field, fields
from types import MappingProxyType
from typing import Any, Callable, Dict, Generic, Mapping, Optional, Type, TypeVar, Union

from pytest_embrace.anno import AnnotationInfo, get_pep_593_values
from pytest_embrace.exc import CaseConfigurationError

CaseType = TypeVar("CaseType")
CoCaseType = TypeVar("CoCaseType", contravariant=True)


@dataclass
class CaseArtifact(Generic[CoCaseType]):
    """Wrap the result of test executions.
    Contains the CaseType object created from the enclosing module.
    Conrains the result returned (or yielded) from the case fixture."""

    case: CoCaseType
    actual_result: Any = None


CaseArtifactType = TypeVar("CaseArtifactType", bound=CaseArtifact, covariant=True)
CaseRunner = Callable[..., Any]


CaseCls = TypeVar("CaseCls")
TUnset = object
UNSET: TUnset = object()


@dataclass
class AttrInfo:
    dc_field: Field
    annotations: Optional[AnnotationInfo] = None
    name: str = field(init=False)

    def __post_init__(self) -> None:
        self.name = self.dc_field.name


@dataclass
class CaseTypeInfo(Generic[CaseCls]):
    type: Type[CaseCls]
    fixture_name: Union[str, TUnset] = UNSET
    type_name: str = field(init=False)
    type_attrs: Dict[str, AttrInfo] = field(default_factory=dict)
    generators: Dict[str, Callable[..., CaseCls]] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.type_name = self.type.__name__
        self.type_annotations = get_pep_593_values(self.type)
        self.type_attrs = {
            f.name: AttrInfo(dc_field=f, annotations=self.type_annotations.get(f.name))
            for f in fields(self.type)
        }


# for some reason, using a dataclass here was problematic.
class Trickle:
    """This type is used to identify attributes marked with trickles()."""

    def __init__(self, no_override: bool = False):
        self.no_override = no_override


OptionalMissing = Optional[_MISSING_TYPE]


def __field(**kwargs: Any) -> Field:
    """Fighting with overload variants of `field()` was so hard.
    This feels like an ok hack for now."""
    return field(**kwargs)


def _kw_only_maybe(kw_only: bool) -> Dict[str, bool]:
    return {} if sys.version_info < (3, 10) else {"kw_only": kw_only}


def trickles(
    *,
    no_override: bool = False,
    default: Any = MISSING,
    default_factory: Union[Callable[[], Any], OptionalMissing] = MISSING,
    init: bool = True,
    repr: bool = True,
    hash: Optional[bool] = None,
    compare: bool = True,
    metadata: Optional[Mapping[Any, Any]] = None,
    kw_only: bool = False,
) -> Any:
    """Marks an attribute as one that can 'trickle down' from module scope
    into table cases as a default value.

    Attempts to remain faithful to the field() API."""
    if metadata is None:
        metadata = MappingProxyType({})

    if default is MISSING and default_factory is MISSING:
        default_ = Trickle(no_override)
    else:
        default_ = default

    field_ = __field(
        default=default_,
        init=init,
        default_factory=default_factory,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata={**metadata, "trickle": Trickle(no_override)},
        **_kw_only_maybe(kw_only),
    )
    return field_


def _pass_thru_parser(extracted: str) -> Any:
    return extracted


def derive_from_filename(
    *,
    pattern: str = r"[\w\.]*test_([\w].*)",
    parse: Callable[[str], Any] = _pass_thru_parser,
    # context: Dict[Any, Any] = None,
    default: Any = MISSING,
    default_factory: Union[Callable[[], Any], OptionalMissing] = MISSING,
    init: bool = True,
    repr: bool = True,
    hash: Optional[bool] = None,
    compare: bool = True,
    metadata: Optional[Mapping[Any, Any]] = None,
    kw_only: bool = False,
) -> Any:
    """Marks an attribute as derived from the name of an enclosing test file."""
    if metadata is None:
        metadata = MappingProxyType({})

    if default is MISSING and default_factory is MISSING:
        default_ = DeriveFromFileName(pattern, parse=parse)
    else:
        default_ = default

    field_: Field = trickles(
        default=default_,
        init=init,
        default_factory=default_factory,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata={
            **metadata,
            "trickle": Trickle(no_override=True),
            "derive_from_filename": default_,
        },
        **_kw_only_maybe(kw_only),
    )
    return field_


class DeriveFromFileName:
    """Extract the value of the annotated attribute from the test's filename
    based on the given regex."""

    def __init__(
        self,
        file_pattern: str = r"[\w\.]*test_([\w].*)",
        *,
        parse: Callable[[str], Any] = _pass_thru_parser,
    ):
        self.file_pattern: re.Pattern = re.compile(file_pattern)
        self.parse = parse

    def get_attr_value(self, filename: str) -> Any:
        match = self.file_pattern.search(filename)
        if match is None:
            raise CaseConfigurationError(
                f"Could not derive a value from filename {filename}"
                f" with regex {self.file_pattern}. :("
            )
        extracted = match.group(1)

        return self.parse(extracted)



from __future__ import annotations

from collections.abc import Iterator
from functools import partial
from inspect import signature
from typing import Any, Callable, Generic
from typing import MutableMapping as TMutableMapping
from typing import NoReturn, Type, TypeVar

import pytest
from typing_extensions import ParamSpec

from .case import CaseArtifact, CaseRunner, CaseType, CaseTypeInfo
from .exc import CaseConfigurationError, TwoStepApiDeprecationError

RegistryValue = Type["CaseType"]

T = TypeVar("T", bound=Type)
TUnset = object
UNSET: TUnset = object()
CodeGenFuncArgs = ParamSpec("CodeGenFuncArgs")
CodeGenFuncReturn = TypeVar("CodeGenFuncReturn")


TCaseInfo = TypeVar("TCaseInfo", bound=CaseTypeInfo)


class CaseTypeRegistry(TMutableMapping[str, TCaseInfo]):
    def __getitem__(self, key: str) -> TCaseInfo:
        return self.__dict__[key]

    def __setitem__(self, key: str, val: TCaseInfo) -> None:
        if key not in self.__dict__:
            self.__dict__[key] = val
            return

        existing = self.__dict__[key]

        if type(existing) == type(val):
            return

        raise CaseConfigurationError(
            f"Already registered a case fixture called for {key} for {val}."
        )

    def __iter__(self) -> Iterator[str]:
        return iter(self.__dict__)

    def __len__(self) -> int:
        return len(self.__dict__)

    def __delitem__(self, _: str) -> None:
        raise NotImplementedError

    def __str__(self) -> str:
        return "\n".join(f"{name}  (via {cls})" for name, cls in self.items())


_registry: CaseTypeRegistry[CaseTypeInfo] = CaseTypeRegistry()


class Embrace(Generic[CaseType]):
    """The entrypoint for creating test cases from a dataclass.
    Register a dataclass as a module test schema and create a configurator for defining
    how tests that implement it will run."""

    def __init__(self, case_type: Type[CaseType]):
        self.case_type = case_type
        self.wrapped_func: CaseRunner | None = None
        self.runner: partial | None = None
        self.generators: dict[str, Callable[..., Any]]
        self.fixture_name: str = ""

    def register_case_runner(self, *_: Any) -> NoReturn:
        raise TwoStepApiDeprecationError(deprecated_method="register_case_runner")

    def caller_fixture_factory(self, *_: Any) -> NoReturn:
        raise TwoStepApiDeprecationError(deprecated_method="caller_fixture_factory")

    def fixture(
        self, func: CaseRunner
    ) -> Callable[[CaseType, pytest.FixtureRequest], Iterator[CaseArtifact[CaseType]]]:
        """Create a fixture to use in Embrace module-based tests.

        The decorated function is a 'virtual' fixture, insofar as its ultimately just
        a plain function. The generated fixture gets parametrized
        (during pytest_generate_tests) with CaseType objects built from enclosing test
        modules and then later calls that decorated function with fixture values."""
        self.fixture_name = func.__name__
        _registry[self.fixture_name] = CaseTypeInfo(
            type=self.case_type,
            fixture_name=func.__name__,
        )

        @pytest.fixture
        def fix(
            case: CaseType,
            request: pytest.FixtureRequest,
        ) -> Iterator[CaseArtifact[CaseType]]:
            """The _real_ fixture that calls the decorated 'virtual' one.

            Uses the FixtureRequest to find values for fixtures based on the virtual
            fixture's parameter names."""
            sig = signature(func)
            kwargs = {name: request.getfixturevalue(name) for name in sig.parameters}
            artifact = CaseArtifact(case=case)
            run_result = func(**kwargs)
            try:
                # fixtures with cleanup yield once
                test_result = next(run_result)
            except TypeError:
                test_result = run_result

            artifact.actual_result = test_result
            yield artifact

            try:
                # trigger the cleanup of the runner, if needed
                next(run_result, ...)
            except TypeError:
                pass

        return fix

    def generator(
        self, func: Callable[CodeGenFuncArgs, CodeGenFuncReturn]
    ) -> Callable[CodeGenFuncArgs, CodeGenFuncReturn]:
        _registry[self.fixture_name].generators[func.__name__] = func

        def render(
            *args: CodeGenFuncArgs.args, **kwargs: CodeGenFuncArgs.kwargs
        ) -> CodeGenFuncReturn:
            case_or_case_iter = func(*args, **kwargs)
            if isinstance(case_or_case_iter, Iterator):
                module_vars = next(case_or_case_iter)
                return module_vars  # TODO!
            else:
                # assert isinstance(case_or_case_iter, CaseType)
                return case_or_case_iter

        return render


def registry() -> CaseTypeRegistry[CaseTypeInfo]:
    return _registry



from __future__ import annotations

from dataclasses import MISSING, asdict, fields, is_dataclass
from types import ModuleType
from typing import Any, Dict, Generic, List, Mapping, Optional, Tuple, Union

import pytest
from pydantic import create_model
from pydantic.error_wrappers import ValidationError
from pydantic.types import StrictBool, StrictBytes, StrictFloat, StrictInt, StrictStr

from pytest_embrace.case import CaseTypeInfo

from .anno import get_pep_593_values
from .case import CaseType, Trickle
from .exc import CaseConfigurationError, EmbraceError

ShouldBecomeStrictBuiltinTypes = Union[str, bytes, int, float, bool]
StrictPydanticTypes = Union[StrictStr, StrictBytes, StrictInt, StrictFloat, StrictBool]

PYDANTIC_STRICTIFICATION_MAP = {
    str: StrictStr,
    bytes: StrictBytes,
    int: StrictInt,
    float: StrictFloat,
    bool: StrictBool,
    # in dataclasses, Field.type is a string at times
    "str": StrictStr,
    "bytes": StrictBytes,
    "int": StrictInt,
    "float": StrictFloat,
    "bool": StrictBool,
}


class ModuleInfo(Generic[CaseType]):
    def __init__(self, *, case_type_info: CaseTypeInfo, module: ModuleType) -> None:
        self.case = case_type_info
        self.module = module
        self.name = module.__name__
        self.cls_fields = {field.name: field for field in fields(self.case.type)}
        self.filename_values = {
            k: derive.get_attr_value(module.__name__)
            for k, v in self.cls_fields.items()
            if (derive := v.metadata.get("derive_from_filename")) is not None
        }
        self.module_attrs = {
            attr: getattr(module, attr)
            for attr in dir(module)
            if not attr.startswith("__") and attr in self.cls_fields
        }
        self.module_attrs.update(self.filename_values)

        self.table = getattr(module, "table", None)
        self.has_table_of_tests = (
            isinstance(self.table, list)
            and len(self.table) > 0
            # would love to use an isinstance() here, but somehow
            # (in Pytester tests specifically)
            # it managed to not work? despite extensive time in PDB?
            and all(str(type(x)) == str(self.case.type) for x in self.table)
        )

        self.pep_539 = get_pep_593_values(self.case.type)

        if self.has_table_of_tests:
            self._trickle_down_table()

    def __str__(self) -> str:
        return f"Module[{self.name}]"

    def to_case(self) -> CaseType:
        kwargs = {k: v for k, v in self.module_attrs.items() if k != "table"}
        try:
            return self.case.type(**kwargs)
        except TypeError:
            raise CaseConfigurationError(
                f"Incorrect or missing attributes in {self.name}."
                f"\nOnly got attributes {kwargs}"
            )

    def _trickle_down_table(self) -> None:
        assert self.table is not None
        UNSET = object()
        trickle_defaults: Dict[str, Tuple[Any, Trickle]] = {
            k: (self.module_attrs.get(k, UNSET), trickle)
            for k, v in self.cls_fields.items()
            if (trickle := v.metadata.get("trickle")) is not None
        }
        all_trickles_unset = len(trickle_defaults) and all(
            v is UNSET for v, _ in trickle_defaults.values()
        )

        for i, case in enumerate(self.table):
            for k, v in asdict(case).items():
                if k in self.filename_values:
                    setattr(case, k, self.filename_values[k])
                elif all_trickles_unset and isinstance(v, Trickle):
                    raise CaseConfigurationError(
                        f"'{k}' is unset at the module level and in table[{i}]:{case}."
                        " '{k}' is marked as 'trickles()' so it must be set somewhere."
                    )
                elif k in trickle_defaults and isinstance(v, Trickle):
                    # absorb the default trickle value
                    (trickle_value, _) = trickle_defaults[k]
                    setattr(case, k, trickle_value)
                elif k in trickle_defaults:
                    # there was a trickle, but it was overridden
                    (_, trickle_config) = trickle_defaults[k]
                    if trickle_config.no_override and k in self.module_attrs:
                        raise CaseConfigurationError(
                            f"Trickle-down attribute '{k}"
                            f" cannot be overridden in table[{i}]:{case}'"
                        )


def load(test: ModuleInfo[CaseType]) -> List[CaseType]:
    if test.table is not None:
        return [
            revalidate_dataclass(case, alias=f"{test}.table[{i}]{case}")
            for i, case in enumerate(test.table)
        ]

    return [revalidate_dataclass(test.to_case(), alias=str(test))]


def _raise_non_dataclass(o: object) -> None:
    if not is_dataclass(o):
        raise CaseConfigurationError("Must use a dataclass for case object.")


def _report_validation_error(exc: ValidationError, *, target_name: str) -> None:
    errors = exc.errors()
    errors_disambiguation = "\n".join(
        f"    Variable/Arg '{e['loc'][0]}'"
        f" should be of type {e['type'].lstrip('type_error.')}"
        for e in errors
    )
    raise CaseConfigurationError(
        f"{len(errors)} invalid attr values in '{target_name}':\n"
        f"{errors_disambiguation}"
    ) from exc


class PydanticConfig:
    arbitrary_types_allowed = True


def revalidate_dataclass(case: CaseType, *, alias: str) -> CaseType:
    _raise_non_dataclass(case)
    kwargs = asdict(case)
    pep539 = get_pep_593_values(type(case))
    validator_kwargs: Dict[str, Tuple[Any, Any]] = {
        field.name: (
            PYDANTIC_STRICTIFICATION_MAP.get(
                pep539[field.name].type if field.name in pep539 else field.type,
                field.type,
            ),
            field.default
            if field.default is not MISSING
            else (
                field.default_factory() if field.default_factory is not MISSING else ...
            ),
        )
        for field in fields(case)
    }

    Validator = create_model(
        f"{case.__class__.__name__}__CaseValidator",
        **validator_kwargs,
        __config__=PydanticConfig,  # type: ignore
    )

    try:
        Validator(**kwargs)
    except ValidationError as e:
        _report_validation_error(e, target_name=alias)

    return case


def find_embrace_requester(
    *, metafunc: pytest.Metafunc, registry: Mapping[str, CaseTypeInfo]
) -> Optional[ModuleInfo]:
    potentials = [name for name in metafunc.fixturenames if name in registry]

    if len(potentials) > 1:
        raise EmbraceError(f"Can't request multiple Embrace fixtures. Got {potentials}")

    found = next(iter(potentials), None)
    return (
        found
        if found is None
        else ModuleInfo(case_type_info=registry[found], module=metafunc.module)
    )



from __future__ import annotations

import pytest

from .utils import make_autouse_conftest, make_test_run_outcome_fixture

_ = make_autouse_conftest(
    """
         from dataclasses import dataclass
         import pytest

         from pytest_embrace import Embrace
         from pytest_embrace.case import CaseArtifact


         @pytest.fixture
         def fix() -> str:
             return "hey"


         @dataclass
         class MyCase:
             name: str


         embrace = Embrace(MyCase)


         @embrace.fixture
         def simple_case(case: MyCase, fix: str) -> None:
             assert case.name == fix
             return {"backwards": [*reversed(case.name)]}
         """
)

single_test_outcome = make_test_run_outcome_fixture(
    """
    name = 'hey'

    def test(simple_case):
        ..."""
)


def test_single(single_test_outcome: pytest.RunResult) -> None:
    single_test_outcome.assert_outcomes(passed=1)


table_test_outcome = make_test_run_outcome_fixture(
    """
    from conftest import MyCase

    table = [MyCase(name='hey'), MyCase(name='yo')]

    def test(simple_case):
        ..."""
)


def test_multi(table_test_outcome: pytest.RunResult) -> None:
    table_test_outcome.assert_outcomes(passed=1, errors=1)


artifact_usage_outcome = make_test_run_outcome_fixture(
    """
    from conftest import MyCase

    name = 'hey'

    def test(simple_case):
        assert simple_case.case.name == 'hey'
        assert simple_case.actual_result['backwards'] == ['y', 'e', 'h']"""
)


def test_use_artifact(artifact_usage_outcome: pytest.RunResult) -> None:
    artifact_usage_outcome.assert_outcomes(passed=1, errors=0)


artifact_plus_fixture_outcome = make_test_run_outcome_fixture(
    """
    from conftest import MyCase

    name = 'hey'

    def test(simple_case, fix):
        assert simple_case.actual_result['backwards'] == list(reversed(fix))"""
)


def test_use_artifact_plus_other(
    artifact_plus_fixture_outcome: pytest.RunResult,
) -> None:
    artifact_plus_fixture_outcome.assert_outcomes(passed=1, errors=0)



from __future__ import annotations

import pytest

from ..utils import generated_module_stdout_factory, make_autouse_conftest

_ = make_autouse_conftest(
    """
from dataclasses import dataclass
import pytest

from pytest_embrace import Embrace
from pytest_embrace.case import CaseArtifact


@dataclass
class CoercionCase:
    number: int


embrace = Embrace(CoercionCase)


@embrace.fixture
def coercion_case(case: CoercionCase) -> None:
    pass


@embrace.generator
def gen(number: int):
    return CoercionCase(number=number)
"""
)


# @pytest.mark.xfail(reason="Type coercion not implemented :(")
def test_dynamic_name_generator(pytester: pytest.Pytester) -> None:
    outcome = pytester.runpytest("--embrace=coercion_case:gen number=4")
    outcome.stdout.fnmatch_lines(
        generated_module_stdout_factory(
            "number = 4",
            case_type="CoercionCase",
            fixture="coercion_case",
        ),
    )



from __future__ import annotations

import pytest

from tests.test_plugin.utils import make_autouse_conftest

_ = make_autouse_conftest(
    """

from dataclasses import dataclass

from pytest_embrace import Embrace, derive_from_filename, trickles


@dataclass
class DeriveAndTableCase:
    horse_name: str
    horse_type: str = derive_from_filename()
    horse_age: int = trickles()


mix_n_match = Embrace(DeriveAndTableCase)


@mix_n_match.fixture
def mix_n_match_case(case: DeriveAndTableCase) -> None:
    if case.horse_name == 'Willybilly':
        assert case.horse_age == 22
        return

    if case.horse_name == 'Jolene':
        assert case.horse_age == 33
        return
"""
)


TEST_FILE = """
from conftest import DeriveAndTableCase


horse_age = 22

table = [
    DeriveAndTableCase(horse_name='Willybilly'),
    DeriveAndTableCase(horse_name='Jolene', horse_age=33),
]


def test(mix_n_match_case):
    assert mix_n_match_case.case.horse_type == 'palomino'
"""


def test_success(pytester: pytest.Pytester) -> None:
    dir = pytester.mkpydir("test_nested")
    file = dir / "test_palomino.py"
    file.touch()
    with file.open("w") as f:
        f.write(TEST_FILE)
    outcome = pytester.runpytest()
    outcome.assert_outcomes(passed=2)



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable

from pytest_embrace import derive_from_filename
from pytest_embrace.loader import load
from tests.conftest import ModuleInfoFactory


@dataclass
class AnnotatedWithDeriveNameCase:
    magic_spell: str = derive_from_filename()


def test_get_name_from_file_default(module_info_factory: ModuleInfoFactory) -> None:
    target = module_info_factory.build(
        AnnotatedWithDeriveNameCase, __name__="my.tests.test_mage_hand"
    )
    (loaded,) = load(target)
    assert loaded.magic_spell == "mage_hand"


class SomeNamesmace:
    def cool_stuff(self) -> str:
        return "cool"


ns = SomeNamesmace()


@dataclass
class DeriveWithParserCase:
    stuff: Callable = derive_from_filename(parse=lambda x: getattr(ns, x))


def test_get_name_from_file_with_parser(module_info_factory: ModuleInfoFactory) -> None:
    target = module_info_factory.build(DeriveWithParserCase, __name__="test_cool_stuff")
    (loaded,) = load(target)
    assert loaded.stuff() == "cool"



from __future__ import annotations

__version__ = "3.0.1"


from .case import CaseArtifact as CaseArtifact
from .case import derive_from_filename as derive_from_filename
from .case import trickles as trickles
from .codegen import RenderModuleBody as RenderModuleBody
from .codegen import RenderText as RenderText
from .embrace import Embrace as Embrace

# defining __all__ for the sake of generating docs with pdoc
__all__ = ["CaseArtifact", "derive_from_filename", "trickles", "Embrace"]



"""Test usage of the `Render()` helper."""
# skip the future import because it breaks dataclass.fields()
# and turns each Field.type to a string!
# isort: dont-add-import: from __future__ import annotations


from __future__ import annotations

import pytest
from pyperclip import sys

from ..utils import generated_module_stdout_factory, make_autouse_conftest

_ = make_autouse_conftest(
    """
from dataclasses import dataclass
import pytest

from pytest_embrace import Embrace, RenderText
from pytest_embrace.case import CaseArtifact


@dataclass
class GreatCase:
    stuff: tuple[int, ...]


embrace = Embrace(GreatCase)


@embrace.fixture
def great(case: GreatCase) -> None:
    pass


@embrace.generator
def gen(how_many: int):
    items = range(how_many)
    inner_tuple = ", ".join(map(str, items))
    return GreatCase(
        stuff=RenderText(f"({inner_tuple})")
    )
"""
)


@pytest.mark.skipif(
    sys.version_info < (3, 9),
    reason="""Many difficulties after adding __future__ imports across the codebase.
    3.8 support is an increasingly low priority, so not bothering with extra logic
    to handle that in this test.""",
)
def test(pytester: pytest.Pytester) -> None:
    outcome = pytester.runpytest("--embrace=great:gen how_many=4")
    outcome.stdout.fnmatch_lines(
        generated_module_stdout_factory(
            "stuff = (0, 1, 2, 3)",
            case_type="GreatCase",
            fixture="great",
        ),
    )



from __future__ import annotations

import traceback
from importlib import reload

import pytest
from pyperclip import copy

from pytest_embrace.codegen import CodeGenManager

from .embrace import registry
from .loader import find_embrace_requester, load


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    sut = find_embrace_requester(metafunc=metafunc, registry=registry())
    if sut is None:
        return
    cases = load(sut)
    metafunc.parametrize("case", cases, ids=[str(c) for c in cases])
    reload(metafunc.module)  # this guarantees the safety of module scope.
    # by reloading at the end of a run, any future references to the just-tested module
    # will encounter it in its 'fresh' state.
    # see ../tests/test_plugin/test_safe_mutable_module_scope.py


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(
        "--embrace",
        help=(
            "Takes the name of an Embrace() caller fixture and generates scaffolding"
            " for a new test module that will use it."
            "\nRun `pytest --embrace-ls` to see all the available caller fixtures."
        ),
    )
    parser.addoption(
        "--embrace-ls",
        help="List all the fixtures created for calling Embrace() test suites.",
        action="store_true",
    )
    parser.addoption(
        "--embrace-gen",
        help="Generate code for some fixture:generator",
    )


STOP_LOOP = object()  # for pytest hooks that stop on the first-non-None-result


def pytest_runtestloop(session: pytest.Session) -> object:
    codegen_directive: str | None = session.config.getoption("--embrace")
    do_ls: bool | None = session.config.getoption("--embrace-ls")

    if codegen_directive is None and do_ls is None:
        return None  # allow normal pytest to happen

    reg = registry()

    if codegen_directive is not None:
        codegen = CodeGenManager(codegen_directive, registry=reg)
        if codegen.case_type is None:
            pytest.exit(
                f"No such fixture '{codegen.fixture_name}'."
                f" Your options are {sorted([*reg])}"
            )
        try:
            copypasta = codegen.render()
        except Exception:
            print(f"Error generating {codegen.generator_name}:")
            # TODO print more when -vv given
            traceback.print_exc(limit=0)
            return STOP_LOOP

        print(f"\nCopying the following output to your clipboard:\n\n{copypasta}")
        copy(copypasta)
        return STOP_LOOP

    if do_ls is True:
        print()
        print(f"The following {len(reg)} Embrace() fixtures are abailable:\n")
        print("\n".join(f"{name}  (via {cls.type_name})" for name, cls in reg.items()))

        return STOP_LOOP

    return None
